@@ -407,7 +407,7 @@ class Trader(object):
 
                 if sig != utils.EXIT and not min(stop_loss, take_profit) <= open_price <= max(stop_loss, take_profit) and e > 0:
                     warn('The deal was opened out of range!')
-                    utils.logger.error('The deal was opened out of range!')
+                    utils.logger.error('(%s) The deal was opened out of range! (candle=%d)', self, e)
                     self.winrate = 0.0
                     self.year_profit = 0.0
                     self.losses = 0
@@ -780,28 +780,28 @@ class Trader(object):
         if open_new_order:
             utils.logger.info('(%s) open trade %s', self, predict)
             if self.trading_on_client:
+                with utils.locker:
+                    if predict == 'Exit':
+                        self.client.exit_last_order()
+                        self.__exit_order__ = True
 
-                if predict == 'Exit':
-                    self.client.exit_last_order()
-                    self.__exit_order__ = True
-
-                else:
-                    moneys = self.client.get_balance(self.ticker.split('/')[1])
-                    ticker_price = self.client.get_ticker_price(self.ticker)
-                    if bet_for_trading_on_client is not np.inf:
-                        bet = bet_for_trading_on_client
                     else:
-                        bet = moneys
-                    if bet > moneys:
-                        bet = moneys
-                    bet /= ticker_price
-                    bet *= utils.RESERVE  # reserve to avoid exchange error
+                        moneys = self.client.get_balance(self.ticker.split('/')[1])
+                        ticker_price = self.client.get_ticker_price(self.ticker)
+                        if bet_for_trading_on_client is not np.inf:
+                            bet = bet_for_trading_on_client
+                        else:
+                            bet = moneys
+                        if bet > moneys:
+                            bet = moneys
+                        bet /= ticker_price
+                        bet *= utils.RESERVE  # reserve to avoid exchange error
 
-                    self.client.exit_last_order()  # exit from previous trade (signal)
-                    self.client.order_create(predict,
-                                             self.ticker,
-                                             bet * self._credit_leverages[-1])  # entry new position
-                    self.__exit_order__ = False
+                        self.client.exit_last_order()  # exit from previous trade (signal)
+                        self.client.order_create(predict,
+                                                 self.ticker,
+                                                 bet * self._credit_leverages[-1])  # entry new position
+                        self.__exit_order__ = False
         utils.logger.debug("(%s) returning prediction", self)
         return {
             'predict': predict,
@@ -872,19 +872,20 @@ class Trader(object):
             while True:
                 if not self.__exit_order__ and time() + wait_sl_tp_checking <= open_time + self._sec_interval:
                     sleep(wait_sl_tp_checking)
-                    price = self.client.get_ticker_price(ticker)
-                    min_ = min(self.__last_stop_loss, self.__last_take_profit)
-                    max_ = max(self.__last_stop_loss, self.__last_take_profit)
-                    utils.logger.debug('(%s) checking SL/TP', self)
-                    if (not (min_ < price < max_)) and prediction["predict"] != 'Exit':
-                        self.__exit_order__ = True
-                        utils.logger.info('(%s) exit trade', self)
-                        index = f'{self.ticker}, {ctime()}'
-                        utils.logger.info("(%s) trading prediction exit in sleeping at %s: %s", self, index, prediction)
-                        if print_out:
-                            print("(%s) trading prediction exit in sleeping at %s: %s" % (self, index, prediction))
-                        if self.trading_on_client:
-                            self.client.exit_last_order()
+                    with utils.locker:
+                        price = self.client.get_ticker_price(ticker)
+                        min_ = min(self.__last_stop_loss, self.__last_take_profit)
+                        max_ = max(self.__last_stop_loss, self.__last_take_profit)
+                        utils.logger.debug('(%s) checking SL/TP', self)
+                        if (not (min_ < price < max_)) and prediction["predict"] != 'Exit':
+                            self.__exit_order__ = True
+                            utils.logger.info('(%s) exit trade', self)
+                            index = f'{self.ticker}, {ctime()}'
+                            utils.logger.info("(%s) trading prediction exit in sleeping at %s: %s", self, index, prediction)
+                            if print_out:
+                                print("(%s) trading prediction exit in sleeping at %s: %s" % (self, index, prediction))
+                            if self.trading_on_client:
+                                self.client.exit_last_order()
                 if time() >= (open_time + self._sec_interval):
                     self._prev_predict = utils.convert_signal_str(self.returns[-1])
                     open_time += self._sec_interval
@@ -937,7 +938,8 @@ class Trader(object):
             def get_trading_predict(self,
                                     bet_for_trading_on_client: Union[float, int] = np.inf,
                                     ) -> Dict[str, Union[str, float]]:
-                balance = self.client.get_balance(self.ticker.split('/')[1])
+                with utils.locker:
+                    balance = self.client.get_balance(self.ticker.split('/')[1])
                 bet = bet_for_trading_on_client_copy
                 if bet_for_trading_on_client is np.inf:
                     if TradingClient.cls_open_orders != can_orders:
@@ -1103,6 +1105,20 @@ class Trader(object):
         self.set_credit_leverages()
         return self.returns
 
+    def correct_sl_tp(self,
+                      sl_correction: Union[float, int] = 50,
+                      tp_correction: Union[float, int] = 50):  # TODO: make documentation for this method   <-----------------------------------------------------------------------------------------------------------------
+        for e, (sl, tp, p, sig) in enumerate(zip(self._stop_losses, self._take_profits, self.df['Close'], self.returns)):
+            if sig == utils.BUY and sl > p:
+                self._stop_losses[e] = p * (1 - sl_correction / 10_000)
+            if sig == utils.SELL and sl < p:
+                self._stop_losses[e] = p * (1 + sl_correction / 10_000)
+
+            if sig == utils.SELL and tp > p:
+                self._stop_losses[e] = p * (1 - tp_correction / 10_000)
+            if sig == utils.BUY and tp < p:
+                self._stop_losses[e] = p * (1 + tp_correction / 10_000)
+
 
 class ExampleStrategies(Trader):
 
@@ -1488,25 +1504,23 @@ class ExampleStrategies(Trader):
                                                              self.df['Close'], **sar_kwargs)
         sardown: np.ndarray = sar.psar_down().values
         sarup: np.ndarray = sar.psar_up().values
-        self._stop_losses = list(sar.psar().values)
-
+        self._stop_losses = sar.psar().values.tolist()
         if plot:
-            for SAR_ in (sarup, sardown):
-                self.fig.plot_line(line=sarup,
-                                   width=utils.SAR_UP_WIDTH,
-                                   color=utils.SAR_UP_COLOR,
-                                   name=utils.SAR_UP_NAME,
-                                   opacity=utils.SAR_UP_ALPHA,
-                                   _row=self.fig.data_row,
-                                   _col=self.fig.data_col)
+            self.fig.plot_line(line=sarup,
+                               width=utils.SAR_UP_WIDTH,
+                               color=utils.SAR_UP_COLOR,
+                               name=utils.SAR_UP_NAME,
+                               opacity=utils.SAR_UP_ALPHA,
+                               _row=self.fig.data_row,
+                               _col=self.fig.data_col)
 
-                self.fig.plot_line(line=sardown,
-                                   width=utils.SAR_DOWN_WIDTH,
-                                   color=utils.SAR_DOWN_COLOR,
-                                   name=utils.SAR_DOWN_NAME,
-                                   opacity=utils.SAR_DOWN_ALPHA,
-                                   _row=self.fig.data_row,
-                                   _col=self.fig.data_col)
+            self.fig.plot_line(line=sardown,
+                               width=utils.SAR_DOWN_WIDTH,
+                               color=utils.SAR_DOWN_COLOR,
+                               name=utils.SAR_DOWN_NAME,
+                               opacity=utils.SAR_DOWN_ALPHA,
+                               _row=self.fig.data_row,
+                               _col=self.fig.data_col)
 
         for price, up, down in zip(
                 list(self.df['Close'].values), list(sarup), list(sardown)):
@@ -1520,6 +1534,7 @@ class ExampleStrategies(Trader):
                 self.returns.append(utils.EXIT)
         self.set_credit_leverages()
         self.set_open_stop_and_take(set_stop=False)
+        self.correct_sl_tp()
         return self.returns
 
     def strategy_macd_histogram_diff(self,
@@ -1631,6 +1646,37 @@ class ExampleStrategies(Trader):
         self.set_credit_leverages()
         return self.returns
 
+    def strategy_bollinger_breakout(self,
+                                    plot: bool = True,
+                                    to_mid: bool = False,
+                                    to_opposite: bool = False,
+                                    *bollinger_args,
+                                    **bollinger_kwargs):
+        self.strategy_bollinger(plot=plot,
+                                to_mid=to_mid,
+                                *bollinger_args,
+                                **bollinger_kwargs)
+        self.inverse_strategy()
+        if to_opposite:
+            bollinger: ta.volatility.BollingerBands = ta.volatility.BollingerBands(self.df['Close'],
+                                                                                   fillna=True,
+                                                                                   *bollinger_args,
+                                                                                   **bollinger_kwargs)
+
+            mid_: pd.Series = bollinger.bollinger_mavg()
+            upper: pd.Series = bollinger.bollinger_hband()
+            lower: pd.Series = bollinger.bollinger_lband()
+            self._stop_losses = []
+            for sig, high, low in zip(self.returns,
+                                      upper,
+                                      lower):
+                if sig == utils.BUY:
+                    self._stop_losses.append(low)
+                else:
+                    self._stop_losses.append(high)
+        self.correct_sl_tp()
+        return self.returns
+
     def strategy_idris(self, points=20):
         self._stop_losses = [np.inf] * 2
         self._take_profits = [np.inf] * 2
